#include <sys/eventfd.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <tlssep-common.h>

tlssep_status_t
tlssep_wait(int fd)
{
	eventfd_t value = 0;
	tlssep_status_t status = TLSSEP_STATUS_OK;

	int ret = eventfd_read(fd, &value);

	if (ret == -1 && EAGAIN == errno) {
		status = TLSSEP_STATUS_AGAIN;
	} else if (ret == -1 && EAGAIN != errno) {
		status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
	}

	return status;
}

tlssep_status_t
tlssep_signal(int fd)
{
	int ret = eventfd_write(fd, 1);
	return ret == -1 ? TLSSEP_STATUS_ERROR_RPC_FAILURE : TLSSEP_STATUS_OK;
}

tlssep_status_t
tlssep_send(int signalfd, int waitfd, void *dest, const void *src, int size, int *bytes_written)
{
	assert(signalfd >= 0);
	assert(waitfd   >= 0);
	assert(NULL     != dest);
	assert(NULL     != src);
	assert(NULL     != bytes_written);

	tlssep_status_t status = TLSSEP_STATUS_OK;
	int remaining          = size;
	*bytes_written         = 0;

	while (0 != remaining) {
		int to_write = remaining > TLSSEP_BUFFER_SIZE ? TLSSEP_BUFFER_SIZE : remaining;

		memcpy(dest, src + *bytes_written, to_write);

		remaining      -= to_write;
		*bytes_written += to_write;

		if (0 != remaining) {
			status = tlssep_signal(signalfd);
			if (TLSSEP_STATUS_OK != status) {
				goto done;
			}

			status = tlssep_wait(waitfd);
			if (TLSSEP_STATUS_OK != status) {
				goto done;
			}
		}
	}

done:
	if (TLSSEP_STATUS_OK == status) {
		status = tlssep_signal(signalfd);
	}

	return status;
}

tlssep_status_t
tlssep_recv(int signalfd, int waitfd, tlssep_status_t (*consumer) (void *dest, int offset, const void *src, int size), void *dest, const void *src, int size, int *bytes_read)
{
	assert(signalfd >= 0);
	assert(waitfd   >= 0);
	assert(NULL     != dest);
	assert(NULL     != src);
	assert(NULL     != bytes_read);

	tlssep_status_t status = TLSSEP_STATUS_OK;
	int remaining          = size;
	*bytes_read            = 0;

	while (0 != remaining) {
		int to_read = remaining > TLSSEP_BUFFER_SIZE ? TLSSEP_BUFFER_SIZE : remaining;

		status = consumer(dest, *bytes_read, src, to_read);
		if (TLSSEP_STATUS_OK != status) {
			goto done;
		}

		remaining   -= to_read;
		*bytes_read += to_read;

		if (0 != remaining) {
			status = tlssep_signal(signalfd);
			if (TLSSEP_STATUS_OK != status) {
				goto done;
			}

			status = tlssep_wait(waitfd);
			if (TLSSEP_STATUS_OK != status) {
				goto done;
			}
		}
	}

done:
	return status;
}
