#include <openssl/conf.h>
#include <openssl/err.h>
#include <openssl/ssl.h>
#include <sys/eventfd.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <tlssep.h>
#include <tlssep-common.h>
#include <tlssep-config.h>
#include <tlssep-private.h>
#include <tlssep-verify.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define DESC_SIZE          1024
#define SELECT_LIMIT       1024 /* Max number of F.D.s in fd_set? */
#define NOTIFICATION_FD1   STDERR_FILENO + 1
#define NOTIFICATION_FD2   STDERR_FILENO + 2
#define OP_SOCKET          STDERR_FILENO + 3
#define FD_TRANSFER_SOCKET STDERR_FILENO + 4

typedef struct desc_t {
	int  tlsFd;
	SSL *ssl;
	int  notificationfd;
	bool blocking_io;
	bool notified;
} desc_t;

fd_set             rfds;
const char        *DEFAULT_CONFIG_PATH = SYSCONFDIR "/tlssep-decorator-" API_VERSION ".cfg";
bool               running             =  true;
desc_t             desc[DESC_SIZE];

/* See the SSL_get_error man page. */
static tlssep_status_t
_ssl_error_to_tlssep_status(int err)
{
	tlssep_status_t status = TLSSEP_STATUS_ERROR_TLS;

	switch (err) {
	case SSL_ERROR_NONE:
		status = TLSSEP_STATUS_OK;
		break;
	case SSL_ERROR_ZERO_RETURN:
		status = TLSSEP_STATUS_ERROR_CLOSED;
		break;
	case SSL_ERROR_SYSCALL:
		status = TLSSEP_STATUS_ERROR_SYSCALL;
		break;
	}

	return status;
}

static void
_reset_desc_entry(unsigned int i)
{
	desc[i].tlsFd                     = -1;
	desc[i].ssl                       =  NULL;
	desc[i].notificationfd            = -1;
	desc[i].blocking_io               =  true;
	desc[i].notified                  =  false;
}

static void
_init_svc(tlssep_args_ret_t *args_ret)
{
	debug("Dec: Entering %s\n", __func__);

	args_ret->ret.status = TLSSEP_STATUS_OK;
	
	SSL_library_init();
	SSL_load_error_strings();

	if (TLSSEP_STATUS_OK != tlssep_signal(NOTIFICATION_FD2)) {
		perror("Failed write to notification file descriptor");
		exit(EXIT_FAILURE);
	}

	debug("Dec: Leaving %s\n", __func__);
}

static int
_verify_null_callback(int preverify_ok, X509_STORE_CTX *ctx)
{
	/* NOTE: we handle verification later, after the connection.
	 * It would be nice to handle it here, but sadly this OpenSSL
	 * callback does not take as an argument an arbitrary state.
	 * This makes it difficult to verify using a novel technique.
	 */
	return true;
}

static tlssep_status_t
_tls_init(const char *cert_path, const char *key_path, const SSL_METHOD *(method_fn) (void), int tlsFd, SSL_CTX **ctx, SSL **ssl)
{
	assert(TLS_client_method == method_fn
	    || TLS_server_method == method_fn);
	assert(NULL != ctx);
	assert(NULL != ssl);

	tlssep_status_t status = TLSSEP_STATUS_OK;
	
	const SSL_METHOD *meth = method_fn();
	assert(NULL != meth);

	if (NULL == *ctx) {
		*ctx = SSL_CTX_new(meth);
		if(NULL == *ctx) {
			status = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
			goto done;
		}

		if (SSL_CTX_set_min_proto_version(*ctx, TLS1_3_VERSION) != 1) {
			status = TLSSEP_STATUS_ERROR;
			goto done;
		}

		bool require_client_certificate = false;
		if (true == tlssep_config_lookup_require_client_certificate(&require_client_certificate)
		 && true == require_client_certificate
		 && TLS_server_method == method_fn) {
			SSL_CTX_set_verify(*ctx, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT, _verify_null_callback);
		}

		if(cert_path && SSL_CTX_use_certificate_file(*ctx, cert_path, SSL_FILETYPE_PEM) != 1) {
			status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERT_PATH;
			goto done;
		}

		if(key_path && SSL_CTX_use_PrivateKey_file(*ctx, key_path, SSL_FILETYPE_PEM) != 1) {
			status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_PRIVKEY_PATH;
			goto done;
		}

		if(key_path && SSL_CTX_check_private_key(*ctx) != 1) {
			status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_PRIVKEY;
			goto done;
		}

		if(SSL_CTX_set_default_verify_paths (*ctx) != 1) {
			status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CA_PATH;
			goto done;
		}
	}

	*ssl = SSL_new(*ctx);
	if(NULL == *ssl) {
		status = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
		goto done;
	}

	int sslFD = SSL_set_fd(*ssl, tlsFd);
	if(0 == sslFD) {
		status = TLSSEP_STATUS_ERROR_DECORATOR_ERROR_SETTING_FD;
		goto done;
	}
	
done:
	if(TLSSEP_STATUS_OK != status) {
		if (NULL != *ctx) {
			SSL_CTX_free(*ctx);
			*ctx = NULL;
		}

		if (NULL != *ssl) {
			SSL_free(*ssl);
			*ssl = NULL;
		}
        } 
	
	return status;
}

// http://blog.varunajayasiri.com/passing-file-descriptors-between-processes-using-sendmsg-and-recvmsg:
static int
recv_file_descriptor(int socket)
{
	int fnval = -1;
	struct msghdr message;
	struct iovec iov[1];
	struct cmsghdr *control_message = NULL;
	char ctrl_buf[CMSG_SPACE(sizeof(int))];
	char data[1];
	int res;

	memset(&message, 0, sizeof(struct msghdr));
	memset(ctrl_buf, 0, CMSG_SPACE(sizeof(int)));

	/* For the dummy data */
	iov[0].iov_base = data;
	iov[0].iov_len  = sizeof(data);

	message.msg_name       = NULL;
	message.msg_namelen    = 0;
	message.msg_control    = ctrl_buf;
	message.msg_controllen = CMSG_SPACE(sizeof(int));
	message.msg_iov        = iov;
	message.msg_iovlen     = 1;

	if((res = recvmsg(socket, &message, 0)) <= 0) {
		return res;
	}

	/* Iterate through header to find if there is a file descriptor. */
	for(control_message = CMSG_FIRSTHDR(&message); control_message != NULL; control_message = CMSG_NXTHDR(&message, control_message)) {
		if((control_message->cmsg_level == SOL_SOCKET) && (control_message->cmsg_type == SCM_RIGHTS)) {
			memcpy(&fnval, CMSG_DATA(control_message), sizeof(fnval));
			goto done;
		}
	}

done:
	return fnval;
}

static void
_connect_or_accept_svc(tlssep_args_ret_t *args_ret, int (*SSL_op) (SSL *ssl))
{
	debug("Dec: Entering %s\n", __func__);

	assert(SSL_op == SSL_connect || SSL_op == SSL_accept);

	static SSL_CTX *ctx   =  NULL;
	int    sslCon         = -1;

	args_ret->ret.status = TLSSEP_STATUS_OK;
	args_ret->ret.id     = -1;

	for (args_ret->ret.id = 0; args_ret->ret.id < DESC_SIZE; args_ret->ret.id++) {
		if (desc[args_ret->ret.id].tlsFd == -1) {
			break;
		}
	}

	if (DESC_SIZE == args_ret->ret.id) {
		args_ret->ret.status = TLSSEP_STATUS_ERROR_DESCRIPTORS_EXHAUSTED;
		goto done;
	}

	assert(NULL == desc[args_ret->ret.id].ssl);

	desc[args_ret->ret.id].tlsFd = recv_file_descriptor(FD_TRANSFER_SOCKET);
	if (desc[args_ret->ret.id].tlsFd <= 0
	 || desc[args_ret->ret.id].tlsFd >= SELECT_LIMIT) {
		args_ret->ret.status = TLSSEP_STATUS_ERROR_TRANSMITTING_FILE_DESCRIPTOR;
		goto done;
	}

	desc[args_ret->ret.id].notificationfd = recv_file_descriptor(FD_TRANSFER_SOCKET);
	if (desc[args_ret->ret.id].notificationfd <= 0) {
		args_ret->ret.status = TLSSEP_STATUS_ERROR_TRANSMITTING_FILE_DESCRIPTOR;
		goto done;
	}

	if (SSL_op == SSL_connect) {
		const char *certpath = NULL;
		const char *privkeypath = NULL;
		bool provide_client_certificate = false;
		if (true == tlssep_config_lookup_provide_client_certificate(&provide_client_certificate)
		 && true == provide_client_certificate) {
			if (false == tlssep_config_lookup_certificate_path(&certpath)) {
				args_ret->ret.status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERT_PATH;
				goto done;
			}
			if (false == tlssep_config_lookup_privkey_path(&privkeypath)) {
				args_ret->ret.status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_PRIVKEY_PATH;
				goto done;
			}
		}
		args_ret->ret.status = _tls_init(certpath, privkeypath, TLS_client_method, desc[args_ret->ret.id].tlsFd, &ctx, &desc[args_ret->ret.id].ssl);
	} else if (SSL_op == SSL_accept) {
		const char *certpath = NULL;
		const char *privkeypath = NULL;
		if (false == tlssep_config_lookup_certificate_path(&certpath)) {
			args_ret->ret.status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERT_PATH;
			goto done;
		}
		if (false == tlssep_config_lookup_privkey_path(&privkeypath)) {
			args_ret->ret.status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_PRIVKEY_PATH;
			goto done;
		}
		args_ret->ret.status = _tls_init(certpath, privkeypath, TLS_server_method, desc[args_ret->ret.id].tlsFd, &ctx, &desc[args_ret->ret.id].ssl);
	} else {
		assert(false);
	}

	if(TLSSEP_STATUS_OK != args_ret->ret.status) {
		goto done;
	}

	sslCon = SSL_op(desc[args_ret->ret.id].ssl);
	if (1 != sslCon) {
		args_ret->ret.status = _ssl_error_to_tlssep_status(SSL_get_error(desc[args_ret->ret.id].ssl, sslCon));
		goto done;
	}

	int engine_mask = 0;
	if(false == tlssep_config_lookup_verification_engines(&engine_mask)) {
		args_ret->ret.status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_VERIFICATION_ENGINE;
		goto done;
	}

	int enough;
	if (false == tlssep_config_lookup_enough(&enough)) {
		/* Fall back on requiring all configured engines. */
		int i;
		enough = 0;
		for (i = 0; i < sizeof(engine_mask) * 8; i++) {
			if (engine_mask & 1 << i) {
				enough++;
			}
		}
	}

	assert(sizeof(args_ret->args.buf) > TLSSEP_MAX_COMMON_NAME_SIZE);
	args_ret->ret.status = tlssep_certificate_ok(desc[args_ret->ret.id].ssl, engine_mask, enough, args_ret->args.buf, true == args_ret->args.size ? args_ret->ret.buf : NULL);
	if(TLSSEP_STATUS_OK != args_ret->ret.status) {
		goto done;
	}

	args_ret->ret.size = strlen(args_ret->ret.buf);

	FD_SET(desc[args_ret->ret.id].tlsFd, &rfds);

done:
	if(TLSSEP_STATUS_OK != args_ret->ret.status) {
		if (-1 == desc[args_ret->ret.id].tlsFd) {
			close(desc[args_ret->ret.id].tlsFd);
		}

		if (-1 == desc[args_ret->ret.id].notificationfd) {
			close(desc[args_ret->ret.id].notificationfd);
		}

		if (NULL != ctx) {
			SSL_CTX_free(ctx);
			ctx = NULL;
		}

		if (args_ret->ret.id < DESC_SIZE && NULL != desc[args_ret->ret.id].ssl) {
			SSL_free(desc[args_ret->ret.id].ssl);
		}

		_reset_desc_entry(args_ret->ret.id);
        }

	if (TLSSEP_STATUS_OK != tlssep_signal(NOTIFICATION_FD2)) {
		perror("Failed write to notification file descriptor");
		exit(EXIT_FAILURE);
	}

	debug("Dec: Leaving %s\n", __func__);
}

static tlssep_status_t
_write_some(void *ssl, int offset, const void *buf, int size)
{
	tlssep_status_t fnval = TLSSEP_STATUS_OK;

	while (true) {
		int bytes_written = SSL_write(ssl, buf, size);
		if (bytes_written == size) {
			break;
		} else if(-1 == bytes_written) {
			int err = SSL_get_error(ssl, bytes_written);
			if (SSL_ERROR_WANT_WRITE == err) {
				continue; /* Try again. */
			} else {
				fnval = _ssl_error_to_tlssep_status(err);
				goto done;
			}
		}
	}

done:
	return fnval;
}

static void
_write_svc(tlssep_args_ret_t *args_ret)
{
	debug("Dec: Entering %s\n", __func__);

	assert(args_ret->args.size > 0);

	int bytes_written = 0;

	if (TLSSEP_STATUS_OK != tlssep_recv(NOTIFICATION_FD2, NOTIFICATION_FD1, _write_some, desc[args_ret->args.id].ssl, args_ret->args.buf, args_ret->args.size, &bytes_written)) {
		perror("Failed to receive data from application");
		exit(EXIT_FAILURE);
	}

	args_ret->ret.size   = bytes_written;
	args_ret->ret.status = TLSSEP_STATUS_OK;

	if (TLSSEP_STATUS_OK != tlssep_signal(NOTIFICATION_FD2)) {
		perror("Failed write to notification file descriptor");
		exit(EXIT_FAILURE);
	}

	debug("Dec: Leaving %s\n", __func__);
}

static void
_read_or_peek_svc(tlssep_args_ret_t *args_ret, int (*SSL_op) (SSL *ssl, void *buf, int num))
{
	debug("Dec: Entering %s\n", __func__);

	assert(SSL_read == SSL_op || SSL_peek == SSL_op);

	int sent = 0;
	int readLen = 0;
	char *buf = NULL;
	args_ret->ret.status = TLSSEP_STATUS_OK;

	assert(args_ret->args.size > 0);

	desc[args_ret->args.id].notified = false;

	if (args_ret->args.size <= TLSSEP_BUFFER_SIZE) {
		buf = args_ret->ret.buf;
	} else {
		buf = malloc(args_ret->args.size);
		if (NULL == buf) {
			args_ret->ret.status = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
			goto done;
		}
	}

	int left_to_read = args_ret->args.size;
	int read_so_far = 0;
	do {
		readLen = SSL_op(desc[args_ret->args.id].ssl, buf + read_so_far, left_to_read);
		if(readLen < 0) {
			if (0 != read_so_far && errno == EAGAIN) {
				break;
			} else if (desc[args_ret->args.id].blocking_io == false && errno == EAGAIN) {
				args_ret->ret.status = TLSSEP_STATUS_AGAIN;
				goto done;
			} else {
				args_ret->ret.status = _ssl_error_to_tlssep_status(SSL_get_error(desc[args_ret->args.id].ssl, readLen));
				goto done;
			}
		}

		left_to_read -= readLen;
		read_so_far  += readLen;
	} while (0 != left_to_read && 0 != readLen && SSL_peek != SSL_op); /* Cannot cheat with peek, else will re-read same data. */

	args_ret->ret.size = read_so_far;

done:
	/* This will merely send notification signal if size == 0 (e.g., error/goto done). */
	if (TLSSEP_STATUS_OK != tlssep_send(NOTIFICATION_FD2, NOTIFICATION_FD1, args_ret->ret.buf, buf, args_ret->ret.size, &sent)) {
		perror("Failed to send data to application");
		exit(EXIT_FAILURE);
	}
	assert(args_ret->ret.size == sent);

	if (buf != NULL && buf != args_ret->ret.buf) {
		free(buf);
	}

	debug("Dec: Leaving %s\n", __func__);
}

static void
_setnonblock_svc(tlssep_args_ret_t *args_ret)
{
	debug("Dec: Entering %s\n", __func__);

	int ret = fcntl(desc[args_ret->args.id].tlsFd, F_SETFL, O_NONBLOCK);
	if (ret != -1) {
		desc[args_ret->args.id].blocking_io = false;
		args_ret->ret.status = TLSSEP_STATUS_OK;
	} else {
		args_ret->ret.status = TLSSEP_STATUS_ERROR_DECORATOR_SET_NONBLOCK;
	}

	if (TLSSEP_STATUS_OK != tlssep_signal(NOTIFICATION_FD2)) {
		perror("Failed write to notification file descriptor");
		exit(EXIT_FAILURE);
	}

	debug("Dec: Leaving %s\n", __func__);
}

static void
_close_svc(tlssep_args_ret_t *args_ret)
{
	debug("Dec: Entering %s\n", __func__);

	args_ret->ret.status = TLSSEP_STATUS_OK;

	if (NULL == desc[args_ret->args.id].ssl) {
		goto done;
	}

	assert(desc[args_ret->args.id].ssl != NULL);

	FD_CLR(desc[args_ret->args.id].tlsFd, &rfds);

	int shutdown = SSL_shutdown(desc[args_ret->args.id].ssl);
	if (0 == shutdown) {
		SSL_shutdown(desc[args_ret->args.id].ssl);
	} else if(shutdown < 0) {
		args_ret->ret.status = _ssl_error_to_tlssep_status(SSL_get_error(desc[args_ret->args.id].ssl, shutdown));
		goto done;
	}

	SSL_free(desc[args_ret->args.id].ssl);

	int result = close(desc[args_ret->args.id].tlsFd);
	if(0 != result) {
		args_ret->ret.status = TLSSEP_STATUS_ERROR_DECORATOR_CLOSE;
		goto done;
	}

	if (-1 != desc[args_ret->args.id].notificationfd) {
		result = close(desc[args_ret->args.id].notificationfd);
		if (0 != result) {
			args_ret->ret.status = TLSSEP_STATUS_ERROR_DECORATOR_CLOSE;
			goto done;
		}
	}

done:
	_reset_desc_entry(args_ret->args.id);

	if (TLSSEP_STATUS_OK != tlssep_signal(NOTIFICATION_FD2)) {
		perror("Failed write to notification file descriptor");
		exit(EXIT_FAILURE);
	}

	debug("Dec: Leaving %s\n", __func__);
}

static void
_terminate_svc(tlssep_args_ret_t *args_ret)
{
	running = false;

	tlssep_config_free();

	if (-1 != OP_SOCKET) {
		int result = close(OP_SOCKET);
		if (0 != result) {
			fprintf(stderr, "%s\n", "Error closing file descriptor transfer socket");
		}
	}

	if (-1 != FD_TRANSFER_SOCKET) {
		int result = close(FD_TRANSFER_SOCKET);
		if (0 != result) {
			fprintf(stderr, "%s\n", "Error closing file descriptor transfer socket");
		}
	}

	exit(EXIT_SUCCESS);
}

static void
_notify_fd_set(fd_set rfds, desc_t *desc)
{
	int i;
	for (i = 0; i < DESC_SIZE; i++) {
		if (desc[i].tlsFd != -1 && FD_ISSET(desc[i].tlsFd, &rfds)) {
			if (desc[i].tlsFd != -1 && false == desc[i].notified) {
				assert(-1 != desc[i].notificationfd);
				if (TLSSEP_STATUS_OK != tlssep_signal(desc[i].notificationfd)) {
					perror("Error writing to notification socket");
					exit(EXIT_FAILURE);
				}
				desc[i].notified = true; /* Do not notify again until after a read_op. */
			}
		}
	}
}

static void
_handle_op(fd_set rfds, tlssep_args_ret_t *args_ret)
{
	if (FD_ISSET(NOTIFICATION_FD1, &rfds)) {
		if (TLSSEP_STATUS_OK != tlssep_wait(NOTIFICATION_FD1)) {
			perror("Failed to read from notification file descriptor (did application forget to terminate decorator?)");
			exit(EXIT_FAILURE);
		}

		switch(args_ret->args.op) {
		case TLSSEP_OP_INIT:
			_init_svc(args_ret);
			break;
		case TLSSEP_OP_CONNECT:
			_connect_or_accept_svc(args_ret, SSL_connect);
			break;
		case TLSSEP_OP_ACCEPT:
			_connect_or_accept_svc(args_ret, SSL_accept);
			break;
		case TLSSEP_OP_WRITE:
			_write_svc(args_ret);
			break;
		case TLSSEP_OP_READ:
			_read_or_peek_svc(args_ret, SSL_read);
			break;
		case TLSSEP_OP_PEEK:
			_read_or_peek_svc(args_ret, SSL_peek);
			break;
		case TLSSEP_OP_SETNONBLOCK:
			_setnonblock_svc(args_ret);
			break;
		case TLSSEP_OP_CLOSE:
			_close_svc(args_ret);
			break;
		case TLSSEP_OP_TERMINATE:
			_terminate_svc(args_ret);
			break;
		default:
			fprintf(stderr, "Received bad operation\n");
			exit(EXIT_FAILURE);
		}
	}
}

static const char *
_get_application_path(void)
{
        char *fnval = NULL;
	char parent_exe[PATH_MAX];
        static char path[PATH_MAX];

	pid_t parent_pid = getppid();
	int ret = snprintf(parent_exe, sizeof(parent_exe), "/proc/%d/exe", parent_pid);
	if (ret < 0 || ret >= sizeof(parent_exe)) {
		goto done;
	}

        ssize_t size = readlink(parent_exe, path, sizeof(path) - 1);
        if (-1 == size) {
                goto done;
        }

        path[size] = 0x00;
        fnval = path;

done:
        return fnval;
}

int
main(int argc, char *argv[])
{
	tlssep_args_ret_t *args_ret = NULL;

	if (argc != 1) {
		fprintf(stderr, "Usage: %s\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	if (-1 == fcntl(NOTIFICATION_FD1, F_GETFD)) {
		fprintf(stderr, "%s", "Did not receive first notification file descriptor from parent.\n");
		exit(EXIT_FAILURE);
	}

	if (-1 == fcntl(NOTIFICATION_FD2, F_GETFD)) {
		fprintf(stderr, "%s", "Did not receive second notification file descriptor from parent.\n");
		exit(EXIT_FAILURE);
	}

	if (-1 == fcntl(OP_SOCKET, F_GETFD)) {
		fprintf(stderr, "%s", "Did not receive shared memory file descriptor from parent.\n");
		exit(EXIT_FAILURE);
	}

	if (-1 == fcntl(FD_TRANSFER_SOCKET, F_GETFD)) {
		fprintf(stderr, "%s", "Did not receive file descriptor transfer socket from parent.\n");
		exit(EXIT_FAILURE);
	}

	const char *application_path = _get_application_path();
        if (NULL == application_path) {
		fprintf(stderr, "%s", "Could not determine parent's path.\n");
		exit(EXIT_FAILURE);
        }

	int i;
	for (i = 0; i < DESC_SIZE; i++) {
		_reset_desc_entry(i);
	}

	/* TODO: Hmm---is it safe to allow the application to set the config. path? */
	char *env_config_path = getenv("LIBTLSSEP_CONFIG_PATH");
	if (false == tlssep_config_init(NULL != env_config_path ? env_config_path : DEFAULT_CONFIG_PATH, application_path)) {
		fprintf(stderr, "Could not read %s\n", NULL != env_config_path ? env_config_path : DEFAULT_CONFIG_PATH);
                exit(EXIT_FAILURE);
	}

	args_ret = (tlssep_args_ret_t *) mmap(NULL, sizeof(tlssep_args_ret_t), PROT_READ | PROT_WRITE, MAP_SHARED, OP_SOCKET, 0);
        if (NULL == args_ret) {
		fprintf(stderr, "%s", "Could not open shared memory.\n");
		exit(EXIT_FAILURE);
        }

	/* Select on:
	 *	(1) RPC fd or RPC fds
	 *	(2) Network socket
	 * Handle RPC and, when network socket ready to read, write to notification 
	 * socket.
	 *
	 * Application can select on notification socket and RPC on RPC socket when
	 * select returns. Applications: DON'T FORGET TO READ BYTE from notification
	 * socket too!
	 */
	while (true == running) {
		int maxFd = OP_SOCKET;	

		FD_ZERO(&rfds);
		FD_SET(NOTIFICATION_FD1, &rfds);

		for (i = 0; i < DESC_SIZE; i++) {
			/* NOTE: do not select on TLS F.D. if application has already been notified. */
			if (desc[i].tlsFd != -1 && false == desc[i].notified) {
				FD_SET(desc[i].tlsFd, &rfds);
				maxFd = desc[i].tlsFd > maxFd ? desc[i].tlsFd : maxFd;
			}
		}

		int count = select(maxFd + 1, &rfds, NULL, NULL, NULL);

		switch (count) {
		case -1:
			perror("Error calling select");
			// exit(EXIT_FAILURE);
		case 0:
			break;
		default:
			_handle_op(rfds, args_ret);
			_notify_fd_set(rfds, desc);
			break;
		}
	}

        exit(EXIT_SUCCESS);
}
