#include <openssl/x509v3.h>
#include <assert.h>
#include <tlssep-verify.h>
#include <string.h>

/* See the X509_STORE_CTX_get_error and verify man pages. */
static tlssep_status_t
_x509_error_to_tlssep_status(long err)
{
	tlssep_status_t status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE;

	switch(err) {
	case X509_V_OK:
		status = TLSSEP_STATUS_OK;
		break;
	case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_MISSING_ISSUER;
		break;
	case X509_V_ERR_UNABLE_TO_GET_CRL:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_MISSING_CRL;
		break;
	case X509_V_ERR_UNABLE_TO_DECRYPT_CERT_SIGNATURE:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_DECRYPT_SIGNATURE;
		break;
	case X509_V_ERR_UNABLE_TO_DECRYPT_CRL_SIGNATURE:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_DECRYPT_CRL_SIGNATURE;
		break;
	case X509_V_ERR_UNABLE_TO_DECODE_ISSUER_PUBLIC_KEY:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_DECODE_ISSUER_PUBLIC_KEY;
		break;
	case X509_V_ERR_CERT_SIGNATURE_FAILURE:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_SIGNATURE;
		break;
	case X509_V_ERR_CRL_SIGNATURE_FAILURE:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_CRL_SIGNATURE;
		break;
	case X509_V_ERR_CERT_NOT_YET_VALID:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_NOT_YET_VALID;
		break;
	case X509_V_ERR_CERT_HAS_EXPIRED:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_EXPIRED;
		break;
	case X509_V_ERR_CRL_NOT_YET_VALID:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_CRL_NOT_YET_VALID;
		break;
	case X509_V_ERR_CRL_HAS_EXPIRED:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_CRL_EXPIRED;
		break;
	case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_INVALID_NOT_BEFORE;
		break;
	case X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_INVALID_NOT_AFTER;
		break;
	case X509_V_ERR_ERROR_IN_CRL_LAST_UPDATE_FIELD:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_INVALID_CRL_LAST_UPDATE;
		break;
	case X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_INVALID_CRL_NEXT_UPDATE;
		break;
	case X509_V_ERR_OUT_OF_MEM:
		status = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
		break;
	case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_SELF_SIGNED;
		break;
	case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_SELF_SIGNED_CHAIN;
		break;
	case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_MISSING_ISSUER;
		break;
	case X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_LEAF;
		break;
	case X509_V_ERR_CERT_CHAIN_TOO_LONG:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_CHAIN_TOO_LONG;
		break;
	case X509_V_ERR_CERT_REVOKED:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_REVOKED;
		break;
	case X509_V_ERR_INVALID_CA:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_INVALID_CA;
		break;
	case X509_V_ERR_PATH_LENGTH_EXCEEDED:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_PATH;
		break;
	case X509_V_ERR_INVALID_PURPOSE:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_PURPOSE;
		break;
	case X509_V_ERR_CERT_UNTRUSTED:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_UNTRUSTED;
		break;
	case X509_V_ERR_CERT_REJECTED:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_REJECTED;
		break;
	case X509_V_ERR_SUBJECT_ISSUER_MISMATCH:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_ISSUER_MISMATCH;
		break;
	case X509_V_ERR_AKID_SKID_MISMATCH:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_AKID_SKID_MISMATCH;
		break;
	case X509_V_ERR_AKID_ISSUER_SERIAL_MISMATCH:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_AKID_ISSUER_SERIAL_MISMATCH;
		break;
	case X509_V_ERR_KEYUSAGE_NO_CERTSIGN:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_SIGN_DISALLOWED;
		break;
	case X509_V_ERR_INVALID_EXTENSION:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_INVALID_EXTENSION;
		break;
	case X509_V_ERR_INVALID_POLICY_EXTENSION:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_INVALID_POLICY_EXTENSION;
		break;
	case X509_V_ERR_NO_EXPLICIT_POLICY:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_NO_EXPLICIT_POLICY;
		break;
	case X509_V_ERR_DIFFERENT_CRL_SCOPE:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_DIFFERENT_CRL_SCOPE;
		break;
	case X509_V_ERR_UNSUPPORTED_EXTENSION_FEATURE:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_UNSUPPORTED_EXTENSION_FEATURE;
		break;
	case X509_V_ERR_PERMITTED_VIOLATION:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_PERMITTED_VIOLATION;
		break;
	case X509_V_ERR_EXCLUDED_VIOLATION:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_EXCLUDED_VIOLATION;
		break;
	case X509_V_ERR_SUBTREE_MINMAX:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_SUBTREE_MINMAX;
		break;
	case X509_V_ERR_UNSUPPORTED_CONSTRAINT_TYPE:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_UNSUPPORTED_CONSTRAINT_TYPE;
		break;
	case X509_V_ERR_UNSUPPORTED_CONSTRAINT_SYNTAX:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_UNSUPPORTED_CONSTRAINT_SYNTAX;
		break;
	case X509_V_ERR_CRL_PATH_VALIDATION_ERROR:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_CRL_PATH_VALIDATION;
		break;
	case X509_V_ERR_APPLICATION_VERIFICATION:
		status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_MISSING_ISSUER;
		break;
	}

	return status;
}

static tlssep_status_t
_verify_self_signed(SSL *ssl, X509 *cert)
{
	X509_STORE_CTX *ctx = NULL;
	X509_STORE *store = NULL;
	tlssep_status_t status = TLSSEP_STATUS_OK;

	if (NULL == cert) {
		status = TLSSEP_STATUS_ERROR_DECORATOR_MISSING_CERTIFICATE;
		goto done;
	}

	/* Based on work from Zakir Durumeric at
	 * https://zakird.com/2013/10/13/certificate-parsing-with-openssl/.
	 * This seems more robust than merely checking for and accepting:
	 * 	X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT
	 * 	or
	 * 	X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN
	 * from SSL_get_verify_result, as SSL_get_verify_result
	 * notes that it can only return a single error even
	 * when multiple errors exist.
	 *
	 * TODO: Do these allocations unnecessarily affect performance?
	 */

	store = X509_STORE_new();
	if (NULL == store) {
		status = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
		goto done;
	}

	STACK_OF(X509) *sk = SSL_get_peer_cert_chain(ssl);
	if (NULL == sk) {
		status = TLSSEP_STATUS_ERROR_DECORATOR_MISSING_CERTIFICATE_CHAIN;
		goto done;
	}

	int num = sk_X509_num(sk);
	if (0 != num) {
		X509 *top = sk_X509_value(sk, num - 1);
		if (NULL == top) {
			status = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
			goto done;
		}

		X509_STORE_add_cert(store, top);
	}

	/* SSL_get_peer_cert_chain does not return end cert when called on
	 * server side! Add it here just to be sure. */
	X509_STORE_add_cert(store, cert);

	ctx = X509_STORE_CTX_new();
	if (NULL == ctx) {
		status = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
		goto done;
	}

	X509_STORE_CTX_init(ctx, store, cert, sk);

	int result = X509_verify_cert(ctx);
	if (1 != result) {
		status = _x509_error_to_tlssep_status(X509_STORE_CTX_get_error(ctx));
		if (TLSSEP_STATUS_OK != status) {
			goto done;
		}
	}

done:
	if (NULL != ctx) {
		X509_STORE_CTX_free(ctx);
	}

	if (NULL != store) {
		X509_STORE_free(store);
	}

	return status;
}

static tlssep_status_t
_verify_ca(SSL *ssl, X509 *cert)
{
	tlssep_status_t status = TLSSEP_STATUS_OK;

	if (NULL == cert) {
		status = TLSSEP_STATUS_ERROR_DECORATOR_MISSING_CERTIFICATE;
		goto done;
	}

	long result = SSL_get_verify_result(ssl);
	status = _x509_error_to_tlssep_status(result);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

done:
	return status;
}

static tlssep_status_t
_extract_common_name(X509 *cert, char *expected_common_name, char *common_name) {
	assert(NULL != cert);
	assert(NULL != expected_common_name);

	static char common_name_scratch[TLSSEP_MAX_COMMON_NAME_SIZE];
	tlssep_status_t status = TLSSEP_STATUS_ERROR_DECORATOR_CERTIFICATE_MISMATCH;
	GENERAL_NAMES *alt_names = NULL;

	if (NULL == common_name) {
		common_name = common_name_scratch;
	}

	common_name[0] = 0x00;

	alt_names = X509_get_ext_d2i (cert, NID_subject_alt_name, NULL, NULL);
	if (NULL != alt_names) {
		int i, count = sk_GENERAL_NAME_num (alt_names);
		for (i = 0; i < count; i++) {
			GENERAL_NAME *name = sk_GENERAL_NAME_value (alt_names, i);
			if (NULL != name) {
				if (GEN_DNS == name->type) {
					unsigned char *name_utf8;
					ASN1_STRING_to_UTF8 (&name_utf8, name->d.dNSName);
					if (strlen ((char *) name_utf8) != ASN1_STRING_length (name->d.dNSName)) {
						/* Register www.google.com[0x00]example.com?
						 * See http://www.blackhat.com/presentations/bh-usa-09/MARLINSPIKE/BHUSA09-Marlinspike-DefeatSSL-PAPER1.pdf
						 */
						status = TLSSEP_STATUS_ERROR_DECORATOR_CERTIFICATE_NAME_CONTAINS_NULL;
						goto done;
					}
					bool ok = ! strcmp(expected_common_name, (char *) name_utf8);
					OPENSSL_free (name_utf8);
					if (true == ok) {
						strcpy(common_name, expected_common_name);
						status = TLSSEP_STATUS_OK;
						goto done;
					}
				}
			}
		}
	}

	X509_NAME *subj = X509_get_subject_name(cert);
	if (NULL == subj) {
		goto done;
	}

	X509_NAME_get_text_by_NID (subj, NID_commonName, common_name, TLSSEP_MAX_COMMON_NAME_SIZE);

	int i = X509_NAME_get_index_by_NID(subj, NID_commonName, 0);
	if (-1 == i) {
		goto done;
	}

	X509_NAME_ENTRY *entry = X509_NAME_get_entry(subj, i);
	if (NULL == entry) {
		goto done;
	}

	ASN1_STRING *data = X509_NAME_ENTRY_get_data(entry);
	if (NULL == data) {
		goto done;
	}

	if (strlen (common_name) != (size_t) ASN1_STRING_length (data)) {
		/* Register www.google.com[0x00]example.com?
		 * See http://www.blackhat.com/presentations/bh-usa-09/MARLINSPIKE/BHUSA09-Marlinspike-DefeatSSL-PAPER1.pdf
		 */
		status = TLSSEP_STATUS_ERROR_DECORATOR_CERTIFICATE_NAME_CONTAINS_NULL;
		goto done;
	}

	if (0 == strcmp(expected_common_name, common_name)) {
		status = TLSSEP_STATUS_OK;
	}

done:
	return status;
}

tlssep_status_t
tlssep_certificate_ok(SSL *ssl, int engine_mask, int enough, char *expected_common_name, char *common_name)
{
	assert(NULL != ssl);
	assert(NULL != expected_common_name); /* "" means no hostname check. */

	int verification_count = 0;
	tlssep_status_t cn_match_status = TLSSEP_STATUS_OK;
	tlssep_status_t status = TLSSEP_STATUS_OK;

	X509 *cert = SSL_get_peer_certificate(ssl);
	if (NULL != cert) {
		cn_match_status = _extract_common_name(cert, expected_common_name, common_name);
	} else {
		cn_match_status = TLSSEP_STATUS_ERROR_DECORATOR_MISSING_CERTIFICATE;
	}

	int i;
	for (i = 0; i < TLSSEP_VERIFICATION_ENGINE_LAST; i++) {
		int engine = 1 << i;
		if (engine_mask & engine) {
			switch (engine) {
			case TLSSEP_VERIFICATION_ENGINE_UNSAFE:
				status = TLSSEP_STATUS_OK;
				break;
			case TLSSEP_VERIFICATION_ENGINE_CA:
				status = _verify_ca(ssl, cert);
				break;
			case TLSSEP_VERIFICATION_ENGINE_SELF_SIGNED:
				status = _verify_self_signed(ssl, cert);
				break;
			default:
				status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_VERIFICATION_ENGINE;
				goto done;
			}

			if (TLSSEP_STATUS_OK != status) {
				continue;
			}

			if (TLSSEP_VERIFICATION_ENGINE_UNSAFE != engine) {
				if (TLSSEP_STATUS_OK != cn_match_status) {
					status = cn_match_status;
					continue;
				}
			}

			verification_count++;
		}
	}

done:
	if (NULL != cert) {
		X509_free(cert);
	}

	if (TLSSEP_STATUS_OK == status) {
		status = verification_count >= enough ? TLSSEP_STATUS_OK : TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE;
	} else {
		status = verification_count >= enough ? TLSSEP_STATUS_OK : status;
	}

	return status;
}
