#include <sys/eventfd.h>
#include <sys/mman.h>
#include <sys/signalfd.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#define _GNU_SOURCE
#define __USE_GNU
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tlssep.h>
#include <tlssep-common.h>
#include <tlssep-private.h>
#include <unistd.h>
#include <wait.h>

#define PROGRAM_NAME "tlssep-decorator"

typedef struct sync_handler_t {
	bool     installed;
	sigset_t mask_old;
	int      sfd;
} sync_handler_t;

static void
_release_context(tlssep_context_t *context)
{
	if (-1 != context->notificationFd1) {
		close(context->notificationFd1);
	}

	if (-1 != context->notificationFd2) {
		close(context->notificationFd2);
	}

	if (-1 != context->opSocket) {
		close(context->opSocket);
	}

	if (-1 != context->fdTransferSocket) {
		close(context->fdTransferSocket);
	}

	if (NULL != context->shm_name) {
		shm_unlink(context->shm_name);
		free(context->shm_name);
	}

	context->notificationFd1      = -1;
	context->notificationFd2      = -1;
	context->opSocket             = -1;
	context->fdTransferSocket     = -1;
	context->decoratorPid         = -1;
}

static tlssep_status_t
_tear_down_decorator(tlssep_context_t *context)
{
	assert(NULL != context);

	tlssep_status_t status = TLSSEP_STATUS_OK;
	int child_status;
	int wait_status;

	if (-1 != context->opSocket) {
		memset(context->args_ret, 0x00, sizeof(tlssep_args_ret_t));
		context->args_ret->args.op = TLSSEP_OP_TERMINATE;

		status = tlssep_signal(context->notificationFd1);
		if (TLSSEP_STATUS_OK != status) {
			goto done;
		}
	}

done:
	if (-1 != context->decoratorPid) {
		wait_status = waitpid(context->decoratorPid, &child_status, 0);

		if (wait_status != context->decoratorPid || child_status != 0) {
			status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_EXIT;
		}
	}

	_release_context(context);

	return status;
}

static int
_shiftFd(int fd, int offset)
{
	int newFd = -1;

	if (fd != STDERR_FILENO + offset) {
		newFd = dup(fd);
		if (newFd < 0) {
			goto done;
		}

		close(fd);
	} else {
		newFd = fd;
	}

done:
	return newFd;
}

static int
_closeFdsExcept(int skip1, int skip2, int skip3, int skip4)
{
	int fnval = false;
	
	int i;
	/* Close all file descriptors except stdin, out, err, and two given sockets. */
	for (i = STDERR_FILENO + 1; i < getdtablesize(); i++) {
		if (i == skip1 || i == skip2 || i == skip3 || i == skip4) {
			continue;
		}

		if (-1 == close(i) && errno != EBADF) {
			/* Some strange problem; file descriptor
			 * valid (! EBADF), but close failed.
			 */
			goto done;
		}
	}

	skip1 = _shiftFd(skip1, 1);
	if (-1 == skip1) {
		goto done;
	}

	skip2 = _shiftFd(skip2, 2);
	if (-1 == skip2) {
		goto done;
	}

	skip3 = _shiftFd(skip3, 3);
	if (-1 == skip3) {
		goto done;
	}

	skip4 = _shiftFd(skip4, 4);
	if (-1 == skip4) {
		goto done;
	}

	assert(STDERR_FILENO + 1 == skip1);
	assert(STDERR_FILENO + 2 == skip2);
	assert(STDERR_FILENO + 3 == skip3);
	assert(STDERR_FILENO + 4 == skip4);

	assert(-1 != fcntl(skip1, F_GETFD));
	assert(-1 != fcntl(skip2, F_GETFD));
	assert(-1 != fcntl(skip3, F_GETFD));
	assert(-1 != fcntl(skip4, F_GETFD));

	fnval = true;

done:
	return fnval;
}

static tlssep_status_t
_fork_decorator(sync_handler_t *handler, tlssep_context_t *context, int childNotificationFd1, int childNotificationFd2, int childOpSocket, int childFdTransferSocket)
{
	tlssep_status_t status = TLSSEP_STATUS_OK;
	struct stat stat_buf;

	if (-1 == stat(LIBEXECDIR "/" PROGRAM_NAME, &stat_buf)) {
		status = TLSSEP_STATUS_ERROR_DECORATOR_MISSING;
		goto done;
	}

	if (0 == S_ISREG(stat_buf.st_mode)) {
		status = TLSSEP_STATUS_ERROR_DECORATOR_NOT_A_FILE;
		goto done;
	}

	if (0 == (stat_buf.st_mode & 0111)) {
		status = TLSSEP_STATUS_ERROR_DECORATOR_WRONG_PERMS;
		goto done;
	}

	context->decoratorPid = fork();
	if (-1 == context->decoratorPid) {
		status = TLSSEP_STATUS_ERROR_DECORATOR_FORK;
		goto done;
	} else if (context->decoratorPid == 0) {
		/* Child. */
		if (false == _closeFdsExcept(childNotificationFd1, childNotificationFd2, childOpSocket, childFdTransferSocket)) {
			fprintf(stderr, "%s\n", "Error closing existing file descriptors");
			exit(EXIT_FAILURE);
		};

		execl(LIBEXECDIR "/" PROGRAM_NAME, PROGRAM_NAME, NULL);
		fprintf(stderr, "Error executing %s\n", LIBEXECDIR "/" PROGRAM_NAME);
		exit(EXIT_FAILURE);
	}

done:
	return status;
}

tlssep_status_t
tlssep_init (tlssep_context_t *context)
{
	debug("App: Entering %s\n", __func__);

	assert(NULL != context);

	int sockpair[2];
	int childFdTransferSocket = -1;
	tlssep_status_t status = TLSSEP_STATUS_OK;

	context->shm_name         = NULL;
	context->notificationFd1  = -1;
	context->notificationFd2  = -1;
	context->opSocket         = -1;
	context->fdTransferSocket = -1;
	context->decoratorPid     = -1;

	context->notificationFd1 = eventfd(0, EFD_SEMAPHORE);
	if (-1 == context->notificationFd1) {
		status = TLSSEP_STATUS_ERROR_NOTIFICATION_FD_FAILED;
		goto done;
	}

	context->notificationFd2 = eventfd(0, EFD_SEMAPHORE);
	if (-1 == context->notificationFd2) {
		status = TLSSEP_STATUS_ERROR_NOTIFICATION_FD_FAILED;
		goto done;
	}

	static int serial_num = 0;
	if (-1 == asprintf(&context->shm_name, "/tlssep-%d-%d", getpid(), serial_num++)) {
		status = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
		goto done;
	}

	context->opSocket = shm_open(context->shm_name, O_CREAT | O_EXCL | O_RDWR, S_IRWXU | S_IRWXG);
	if (-1 == context->opSocket) {
		status = TLSSEP_STATUS_ERROR_SHM_FAILED;
		goto done;
	}

	int flags = fcntl(context->opSocket, F_GETFD);
	if (-1 == flags) {
		status = TLSSEP_STATUS_ERROR_SHM_FAILED;
		goto done;
        }
	flags &= ~FD_CLOEXEC;
	if (-1 == fcntl(context->opSocket, F_SETFD, flags)) {
		status = TLSSEP_STATUS_ERROR_SHM_FAILED;
		goto done;
	}

	if (-1 == ftruncate(context->opSocket, sizeof(tlssep_args_ret_t))) {
		status = TLSSEP_STATUS_ERROR_SHM_FAILED;
		goto done;
	}

	if (-1 == socketpair(AF_UNIX, SOCK_STREAM, 0, sockpair)) {
		status = TLSSEP_STATUS_ERROR_SOCKET_FAILED;
		goto done;
	}

	context->fdTransferSocket = sockpair[0];
	childFdTransferSocket     = sockpair[1];

	sync_handler_t sync_handler;
	status = _fork_decorator(&sync_handler, context, context->notificationFd1, context->notificationFd2, context->opSocket, childFdTransferSocket);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	context->args_ret = (tlssep_args_ret_t *) mmap(NULL, sizeof(tlssep_args_ret_t), PROT_READ | PROT_WRITE, MAP_SHARED, context->opSocket, 0);
	if (NULL == context->args_ret) {
		status = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
		goto done;
	}

	memset(context->args_ret, 0x00, sizeof(tlssep_args_ret_t));
	context->args_ret->args.op = TLSSEP_OP_INIT;

	status = tlssep_signal(context->notificationFd1);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	status = tlssep_wait(context->notificationFd2);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	status = context->args_ret->ret.status;

done:
	if (-1 != childFdTransferSocket) {
		close(childFdTransferSocket);
	}

	if (NULL == context->args_ret || TLSSEP_STATUS_OK != context->args_ret->ret.status) {
		if (-1 != context->decoratorPid) {
			kill(context->decoratorPid, SIGKILL); /* Make sure child is dead. */
		}
		_release_context(context);
	}

	debug("App: Leaving %s\n", __func__);

	return status;
}

// http://blog.varunajayasiri.com/passing-file-descriptors-between-processes-using-sendmsg-and-recvmsg:
static int
sendFileDescriptor(int socket, int fd_to_send)
{
	struct msghdr message;
	struct iovec iov[1];
	struct cmsghdr *control_message = NULL;
	char ctrl_buf[CMSG_SPACE(sizeof(int))];
	char data[1];

	memset(&message, 0, sizeof(struct msghdr));
	memset(ctrl_buf, 0, CMSG_SPACE(sizeof(int)));

	/* We are passing at least one byte of data so that recvmsg() will not return 0 */
	data[0]         = 'x';
	iov[0].iov_base = data;
	iov[0].iov_len  = sizeof(data);

	message.msg_name       = NULL;
	message.msg_namelen    = 0;
	message.msg_iov        = iov;
	message.msg_iovlen     = 1;
	message.msg_controllen = CMSG_SPACE(sizeof(int));
	message.msg_control    = ctrl_buf;

	control_message             = CMSG_FIRSTHDR(&message);
	control_message->cmsg_level = SOL_SOCKET;
	control_message->cmsg_type  = SCM_RIGHTS;
	control_message->cmsg_len   = CMSG_LEN(sizeof(int));

	memcpy(CMSG_DATA(control_message), &fd_to_send, sizeof(fd_to_send));

	return sendmsg(socket, &message, 0);
}

static tlssep_status_t
_tlssep_connect_or_accept (tlssep_op_t op,
                           tlssep_desc_t *desc,
                           int file_descriptor,
                           const char *expected_name,
                           char *name)
{
	debug("App: Entering %s\n", __func__);

	assert(TLSSEP_OP_CONNECT == op || TLSSEP_OP_ACCEPT == op);
	assert(NULL != desc);
	assert(file_descriptor >= 0);
	assert(NULL != expected_name || TLSSEP_OP_ACCEPT == op);

	tlssep_status_t status = TLSSEP_STATUS_OK;

	memset(desc->context->args_ret, 0x00, sizeof(tlssep_args_ret_t));
        desc->context->args_ret->args.op = op;
	if (NULL != expected_name) {
		assert(MAXHOSTNAMELEN < sizeof(desc->context->args_ret->args.buf));
		strncpy(desc->context->args_ret->args.buf, expected_name, MAXHOSTNAMELEN);
	} else {
		desc->context->args_ret->args.buf[0] = 0x00;
	}
	desc->context->args_ret->args.size = NULL == name ? false : true;

	if (-1 == sendFileDescriptor(desc->context->fdTransferSocket, file_descriptor)) {
		status = TLSSEP_STATUS_ERROR_TRANSMITTING_FILE_DESCRIPTOR;
		goto done;
	}

	close(file_descriptor);

	desc->notificationfd = eventfd(0, EFD_NONBLOCK);
	if (-1 == desc->notificationfd) {
		status = TLSSEP_STATUS_ERROR_NOTIFICATION_FD_FAILED;
		goto done;
	}

	if (-1 == sendFileDescriptor(desc->context->fdTransferSocket, desc->notificationfd)) {
		status = TLSSEP_STATUS_ERROR_TRANSMITTING_FILE_DESCRIPTOR;
		goto done;
	}

	status = tlssep_signal(desc->context->notificationFd1);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	status = tlssep_wait(desc->context->notificationFd2);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	assert(name != NULL || desc->context->args_ret->args.size == false);

	if (NULL != name) {
		if (0 != desc->context->args_ret->ret.size) {
			strcpy(name, desc->context->args_ret->ret.buf);
		} else {
			name[0] = 0x00;
		}
	}

	status   = desc->context->args_ret->ret.status;
	desc->id = desc->context->args_ret->ret.id;

done:
	if (TLSSEP_STATUS_OK != status) {
		if (-1 != desc->notificationfd) {
			close(desc->notificationfd);
		}
	}

	debug("App: Leaving %s\n", __func__);

	return status;
}

tlssep_status_t
tlssep_connect (tlssep_context_t *context,
                int file_descriptor,
                const char *expected_name,
                char *name,
                tlssep_desc_t *desc)
{
	assert(NULL != desc);

	desc->context = context;
	return _tlssep_connect_or_accept(TLSSEP_OP_CONNECT, desc, file_descriptor, expected_name, name);
}

tlssep_status_t
tlssep_accept(tlssep_context_t *context,
              int file_descriptor,
              const char *expected_name,
              char *name,
              tlssep_desc_t *desc)
{
	assert(NULL != desc);

	desc->context = context;
	return _tlssep_connect_or_accept(TLSSEP_OP_ACCEPT, desc, file_descriptor, expected_name, name);
}

tlssep_status_t
tlssep_write (tlssep_desc_t *desc, const void *buf, int buf_size, int *bytes_written)
{
	debug("App: Entering %s\n", __func__);

	assert(NULL != desc);
	assert(NULL != buf);
	assert(NULL != bytes_written);

	tlssep_status_t status = TLSSEP_STATUS_OK;

	if (0 == buf_size) {
		goto done;
	}

	memset(desc->context->args_ret, 0x00, sizeof(tlssep_args_ret_t));
	desc->context->args_ret->args.op = TLSSEP_OP_WRITE;
	desc->context->args_ret->args.id = desc->id;
	desc->context->args_ret->args.size = buf_size;

	status = tlssep_send(desc->context->notificationFd1, desc->context->notificationFd2, desc->context->args_ret->args.buf, buf, buf_size, bytes_written);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	status = tlssep_wait(desc->context->notificationFd2);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

done:
	status         = desc->context->args_ret->ret.status;
	*bytes_written = desc->context->args_ret->ret.size;

	debug("App: Leaving %s\n", __func__);

	return status;
}

/* Consume all event signals from desc->notificationfd. Set ready to
 * true if at least one read succeeded. Both success(es) and EAGAIN
 * return TLSSEP_STATUS_OK.
 */
static tlssep_status_t
_read_ready(tlssep_desc_t *desc, bool *ready)
{
	tlssep_status_t status = TLSSEP_STATUS_OK;

	if (NULL != ready) {
		*ready = false;
	}

	/* Read from notification file descriptor. This file descriptor does
	 * not block, so it is okay if there is no data.
	 */
	while (TLSSEP_STATUS_OK == (status = tlssep_wait(desc->notificationfd))) {
		if (NULL != ready) {
			*ready = true;
		}
	}

	status = TLSSEP_STATUS_AGAIN == status ? TLSSEP_STATUS_OK : status;

	return status;
}

static tlssep_status_t
_memcpy_wrapper(void *dest, int offset, const void *src, int size)
{
	memcpy(dest + offset, src, size);
	return TLSSEP_STATUS_OK;
}

static tlssep_status_t
tlssep_read_op (tlssep_op_t read_op, tlssep_desc_t *desc, void *buf, int buf_size, int *bytes_read)
{
	debug("App: Entering %s\n", __func__);

	assert(TLSSEP_OP_READ == read_op || TLSSEP_OP_PEEK == read_op);
	assert(NULL != desc);
	assert(NULL != buf);
	assert(NULL != bytes_read);

	tlssep_status_t status = TLSSEP_STATUS_OK;

	status = _read_ready(desc, NULL);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	if (buf_size == 0) {
		goto done;
	}

	desc->context->args_ret->args.op   = read_op;
	desc->context->args_ret->args.id   = desc->id;
	desc->context->args_ret->args.size = buf_size;

	status = tlssep_signal(desc->context->notificationFd1);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	status = tlssep_wait(desc->context->notificationFd2);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	status = tlssep_recv(desc->context->notificationFd1, desc->context->notificationFd2, _memcpy_wrapper, buf, desc->context->args_ret->ret.buf, desc->context->args_ret->ret.size, bytes_read);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	status = desc->context->args_ret->ret.status;
	if(TLSSEP_STATUS_OK != status) {
		goto done;
	}

done:
	debug("App: Leaving %s\n", __func__);

	return status;
}

tlssep_status_t
tlssep_read (tlssep_desc_t *desc, void *buf, int buf_size, int *bytes_read)
{
	assert(NULL != desc);
	assert(NULL != buf);
	assert(NULL != bytes_read);

	return tlssep_read_op(TLSSEP_OP_READ, desc, buf, buf_size, bytes_read);
}

tlssep_status_t
tlssep_peek (tlssep_desc_t *desc, void *buf, int buf_size, int *bytes_read)
{
	assert(NULL != desc);
	assert(NULL != buf);
	assert(NULL != bytes_read);

	return tlssep_read_op(TLSSEP_OP_PEEK, desc, buf, buf_size, bytes_read);
}

tlssep_status_t
tlssep_poll (tlssep_desc_t *desc, unsigned int timeout)
{
	debug("App: Entering %s\n", __func__);

	assert(NULL != desc);
	assert(-1   != desc->notificationfd);

	tlssep_status_t status = TLSSEP_STATUS_OK;

	bool ready;
	status = _read_ready(desc, &ready);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	if (true == ready) {
		goto done;
	}

	fd_set rfds, wfds;
	struct timeval tv;
	int retval;

	FD_ZERO(&rfds);
        FD_SET(desc->notificationfd, &rfds);

	FD_ZERO(&wfds);
        FD_SET(desc->notificationfd, &wfds);

	tv.tv_sec  = timeout;
        tv.tv_usec = 0;

	/* FIXME: Change API to support poll read and/or write; see use in wget. */
        retval = select(desc->notificationfd + 1, &rfds, &wfds, NULL, &tv);

	status = retval != -1 ? TLSSEP_STATUS_OK : TLSSEP_STATUS_ERROR_POLL_FAILED;

done:
	debug("App: Leaving %s\n", __func__);

	return status;
}

tlssep_status_t
tlssep_setnonblock (tlssep_desc_t *desc)
{
	debug("App: Entering %s\n", __func__);

	assert(NULL != desc);

	tlssep_status_t status = TLSSEP_STATUS_OK;
	desc->context->args_ret->args.op = TLSSEP_OP_SETNONBLOCK;
	desc->context->args_ret->args.id = desc->id;

	status = tlssep_signal(desc->context->notificationFd1);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	status = tlssep_wait(desc->context->notificationFd2);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	status = desc->context->args_ret->ret.status;

done:
	debug("App: Leaving %s\n", __func__);

	return status;
}

tlssep_status_t
tlssep_close (tlssep_desc_t *desc)
{
	debug("App: Entering %s\n", __func__);

	assert(NULL != desc);

	memset(desc->context->args_ret, 0x00, sizeof(tlssep_args_ret_t));
	desc->context->args_ret->args.op = TLSSEP_OP_CLOSE;
	desc->context->args_ret->args.id = desc->id;
	tlssep_status_t status = TLSSEP_STATUS_OK;

	status = tlssep_signal(desc->context->notificationFd1);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	status = tlssep_wait(desc->context->notificationFd2);
	if (TLSSEP_STATUS_OK != status) {
		goto done;
	}

	status = desc->context->args_ret->ret.status;

done:
	if (-1 != desc->notificationfd) {
		close(desc->notificationfd);
	}

	debug("App: Leaving %s\n", __func__);

	return status;
}

tlssep_status_t
tlssep_terminate (tlssep_context_t *context)
{
	assert(NULL != context);

	return _tear_down_decorator(context);
}

#include <tlssep-status-private.h>

char *
tlssep_strerror (tlssep_status_t code)
{
	int num_errors = sizeof(error_message) / sizeof(error_message_t) - 1;

	int i;
	for (i = 0; i < num_errors; i++) {
		if (error_message[i].code == code) {
			break;
		}
	}

	return error_message[i].message;
}
