#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <assert.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netdb.h>
#include <tlssep.h>
#include <test-api-common.h>

#define NUM_CONNECTIONS 500

const char *hostname = "localhost";

void
testConnect(tlssep_context_t *context, tlssep_desc_t *desc, tlssep_status_t expected, const char *expected_cn)
{
	//int fd = open("/dev/null", 0);
	char cn[TLSSEP_MAX_COMMON_NAME_SIZE];
	const int port = 1025;
	int sock = socket(AF_INET, SOCK_STREAM, 0);
        if(sock < 0) {
                perror("Creating remote socket failed");
                exit (EXIT_FAILURE);
        }
	struct hostent *hp = gethostbyname (hostname);
        if (hp == NULL || hp->h_addr_list[0] == NULL) {
                fprintf (stderr, "Error resolving %s\n", hostname);
                exit (EXIT_FAILURE);
        }

        struct sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	memcpy (&(addr.sin_addr.s_addr), hp->h_addr_list[0], hp->h_length);
	
	int res = connect(sock, (struct sockaddr*)&addr, sizeof(addr));
	if(res < 0) {
		perror("Error connecting to remote host");
		exit(EXIT_FAILURE);
	}

	tlssep_status_t status = tlssep_connect(context, sock, hostname, cn, desc);
	if (expected != status) {
		fprintf(stderr, "Connect test failed: got %s, expected %s\n", tlssep_strerror(status), tlssep_strerror(expected));
		exit(EXIT_FAILURE);
	}

	if (NULL != expected_cn) {
		assert(0 == strcmp(expected_cn, cn));
	}
}

void
testErrors(tlssep_desc_t *desc)
{
	assert(0 == strcmp("No error", tlssep_strerror(TLSSEP_STATUS_OK)));
	assert(0 == strcmp("Generic error", tlssep_strerror(TLSSEP_STATUS_ERROR)));
	assert(0 == strcmp("Bad error code", tlssep_strerror(TLSSEP_STATUS_LAST)));
	assert(0 == strcmp("Bad error code", tlssep_strerror(~0)));
}

int main(int argc, char *argv[])
{
	if (argc > 1) {
		fprintf(stderr, "Usage: %s\n", argv[0]);
                exit(EXIT_FAILURE);
	}

	setenv("LIBTLSSEP_CONFIG_PATH", "tlssep-decorator-self-signed.cfg", true);

	tlssep_context_t context;
	tlssep_desc_t desc[NUM_CONNECTIONS];

	printf("Test first connection ...\n");
	testInit(&context, TLSSEP_STATUS_OK);
	printf("  ... init done\n");
	testConnect(&context, &desc[0], TLSSEP_STATUS_OK, hostname);
	printf("  ... connect done\n");
	testWrite(&desc[0], TLSSEP_STATUS_OK);
	printf("  ... write done\n");
	testPeek(&desc[0], TLSSEP_STATUS_OK);
	printf("  ... peek done\n");
	testRead(&desc[0], TLSSEP_STATUS_OK);
	printf("  ... read done\n");
	testReadRemoteClose(&desc[0], TLSSEP_STATUS_OK);
	printf("  ... read remote close done\n");
	testClose(&desc[0], TLSSEP_STATUS_OK);
	printf("  ... close done\n");

	// Connect and try a second time.
	printf("Test second connection ...\n");
	testConnect(&context, &desc[0], TLSSEP_STATUS_OK, hostname);
	testWrite(&desc[0], TLSSEP_STATUS_OK);
	testRead(&desc[0], TLSSEP_STATUS_OK);
	testClose(&desc[0], TLSSEP_STATUS_OK);

	testTerminate(&context);

	/* Ensure expired certificates fail (server-side chooses expired cert.). */
	printf("Test expired certificate ...\n");
	setenv("LIBTLSSEP_CONFIG_PATH", "tlssep-decorator-expired.cfg", true);
	testInit(&context, TLSSEP_STATUS_OK);
	testConnect(&context, &desc[0], TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_EXPIRED, NULL);
	testTerminate(&context);

	/* Ensure clients can connect without a certificate if permitted. */
	printf("Test no client certificate ...\n");
	setenv("LIBTLSSEP_CONFIG_PATH", "tlssep-decorator-no-client-cert.cfg", true);
	testInit(&context, TLSSEP_STATUS_OK);
	testConnect(&context, &desc[0], TLSSEP_STATUS_OK, hostname);
	testClose(&desc[0], TLSSEP_STATUS_OK);
	testTerminate(&context);

	/* Ensure clients can connect using only global configuration. */
	printf("Test no global configuration ...\n");
	setenv("LIBTLSSEP_CONFIG_PATH", "tlssep-decorator-global.cfg", true);
	testInit(&context, TLSSEP_STATUS_OK);
	testConnect(&context, &desc[0], TLSSEP_STATUS_OK, hostname);
	testClose(&desc[0], TLSSEP_STATUS_OK);
	testTerminate(&context);

	// Connect and try several times concurrently.
	printf("Test multiple connections ... ");
	fflush(stdout);
	testInit(&context, TLSSEP_STATUS_OK);
	int i;
	for (i = 0; i < NUM_CONNECTIONS; i++) {
		testConnect(&context, &desc[i], TLSSEP_STATUS_OK, hostname);
		printf("%d ", i);
		fflush(stdout);
	}
	for (i = NUM_CONNECTIONS - 1; i >= 0; i--) {
		testWrite(&desc[i], TLSSEP_STATUS_OK);
		printf("%d ", i);
		fflush(stdout);
	}
	for (i = 0; i < NUM_CONNECTIONS; i++) {
		testRead(&desc[i], TLSSEP_STATUS_OK);
		testClose(&desc[i], TLSSEP_STATUS_OK);
		printf("%d ", i);
		fflush(stdout);
	}
	testTerminate(&context);
	printf("\n");

	// Test errors.
	printf("Test error codes ...\n");
	testErrors(&desc[0]);

	printf("All tests passed.\n");
		
	exit(EXIT_SUCCESS);
}
