#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <assert.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netdb.h>
#include <tlssep.h>
#include <test-api-common.h>

#define NUM_CONNECTIONS 500

const char *hostname = "localhost"; /* Should be client common name. */
const int PORT = 1025;

int
setupSocket(void)
{
	int server_socket, bind_serv, listen_serv;
        struct sockaddr_in addr;

        server_socket = socket(AF_INET, SOCK_STREAM, 0);
        if(server_socket < 0) {
		perror("Creating socket failed: ");
		exit(EXIT_FAILURE);
        }

	int yes=1;
        if (setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
                perror("Error setting SO_REUSEADDR: ");
                exit(EXIT_FAILURE);
        }

	bzero(&addr, sizeof(addr));
        addr.sin_family = AF_INET;
        addr.sin_port = htons(PORT);
        addr.sin_addr.s_addr = INADDR_ANY;

	bind_serv = bind(server_socket, (struct sockaddr*)&addr, sizeof(addr));
        if(bind_serv != 0) {
                perror("Error binding to port: ");
                exit(EXIT_FAILURE);
        }

	listen_serv = listen(server_socket,0);
        if(listen_serv != 0) {
                perror("Can't configure listening port: ");
                exit(EXIT_FAILURE);
        }

	return server_socket;
}

void
testAccept(tlssep_context_t *context, int listening_socket, tlssep_desc_t *desc, tlssep_status_t expected, const char *expected_cn)
{
	char cn[TLSSEP_MAX_COMMON_NAME_SIZE];
	struct sockaddr_in remote_addr;
	socklen_t socklen = sizeof(remote_addr);

	int client = accept(listening_socket, (struct sockaddr*)&remote_addr, &socklen);
	if(client < 0) {
		perror("Error accepting connection");
		exit(EXIT_FAILURE);
	}

	tlssep_status_t status = tlssep_accept(context, client, hostname, cn, desc);
        if (expected != status) {
		fprintf(stderr, "Accept test failed: got %s, expected %s\n", tlssep_strerror(status), tlssep_strerror(expected));
                exit(EXIT_FAILURE);
        }

	if (NULL != expected_cn) {
		assert(0 == strcmp(expected_cn, cn));
	}
}

int main(int argc, char *argv[])
{
	if (argc > 1) {
		fprintf(stderr, "Usage: %s\n", argv[0]);
                exit(EXIT_FAILURE);
	}

	setenv("LIBTLSSEP_CONFIG_PATH", "tlssep-decorator-self-signed.cfg", true);

	tlssep_context_t context;
	tlssep_desc_t desc[NUM_CONNECTIONS];

	int listeningSocket = setupSocket();
	assert(-1 != listeningSocket);

	printf("Test first connection ...\n");
	testInit(&context, TLSSEP_STATUS_OK);
	printf("  ... init done\n");
	testAccept(&context, listeningSocket, &desc[0], TLSSEP_STATUS_OK, hostname);
	printf("  ... accept done\n");
	testRead(&desc[0], TLSSEP_STATUS_OK);
	printf("  ... read done\n");
	testWrite(&desc[0], TLSSEP_STATUS_OK);
	printf("  ... write done\n");
	testClose(&desc[0], TLSSEP_STATUS_OK);
	printf("  ... close done\n");

	printf("Test second connection ...\n");
	testAccept(&context, listeningSocket, &desc[0], TLSSEP_STATUS_OK, hostname);
	testRead(&desc[0], TLSSEP_STATUS_OK);
	testWrite(&desc[0], TLSSEP_STATUS_OK);
	testClose(&desc[0], TLSSEP_STATUS_OK);

	testTerminate(&context);

	/* Ensure expired certificates fail. */
	printf("Test expired certificate ...\n");
        setenv("LIBTLSSEP_CONFIG_PATH", "tlssep-decorator-expired.cfg", true);
        testInit(&context, TLSSEP_STATUS_OK);
	testAccept(&context, listeningSocket, &desc[0], TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERTIFICATE_EXPIRED, NULL);
        testTerminate(&context);

	/* Ensure clients can connect without a certificate if permitted. */
	printf("Test no client certificate ...\n");
	setenv("LIBTLSSEP_CONFIG_PATH", "tlssep-decorator-no-client-cert.cfg", true);
	testInit(&context, TLSSEP_STATUS_OK);
	testAccept(&context, listeningSocket, &desc[0], TLSSEP_STATUS_OK, "");
	testClose(&desc[0], TLSSEP_STATUS_OK);
	testTerminate(&context);

	/* Ensure clients can connect using only global configuration. */
	printf("Test global configuration ...\n");
	setenv("LIBTLSSEP_CONFIG_PATH", "tlssep-decorator-global.cfg", true);
	testInit(&context, TLSSEP_STATUS_OK);
	testAccept(&context, listeningSocket, &desc[0], TLSSEP_STATUS_OK, "");
	testClose(&desc[0], TLSSEP_STATUS_OK);
	testTerminate(&context);

	// Connect and try several times concurrently.
	printf("Test multiple connections ... ");
        fflush(stdout);
	testInit(&context, TLSSEP_STATUS_OK);
	int i;
	for (i = 0; i < NUM_CONNECTIONS; i++) {
		testAccept(&context, listeningSocket, &desc[i], TLSSEP_STATUS_OK, "");
		printf("%d ", i);
                fflush(stdout);
	}
	for (i = NUM_CONNECTIONS - 1; i >= 0; i--) {
		testRead(&desc[i], TLSSEP_STATUS_OK);
		printf("%d ", i);
                fflush(stdout);
	}
	for (i = 0; i < NUM_CONNECTIONS; i++) {
		testWrite(&desc[i], TLSSEP_STATUS_OK);
		testClose(&desc[i], TLSSEP_STATUS_OK);
		printf("%d ", i);
                fflush(stdout);
	}
	testTerminate(&context);
	printf("\n");

	printf("All tests passed.\n");
		
	exit(EXIT_SUCCESS);
}
